/* $Id: path.c,v 1.3 1998/02/19 01:58:15 svmcguir Exp $ */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "path.h"
#include "util.h"
#include "RANLIB/ranlib.h"

void pathDataAllocate(pathData *pd){
  /* Allocates memory for arrays in pathData structure based on the value
   * of pd->underliers, pd->times and pd->paths.
   */

  allocate2dDouble( &(pd->covariance),   pd->underliers, pd->underliers);
  allocate1dDouble( &(pd->growthRate),   pd->underliers);
  allocate1dDouble( &(pd->dividendRate), pd->underliers);
  allocate1dFloat(  &(pd->cov1),         pd->underliers*pd->underliers);
  allocate1dFloat(  &(pd->fgrowth),      pd->underliers);
  allocate1dFloat(  &(pd->parm),         (pd->underliers*(pd->underliers+3)/2 +1));
  allocate3dDouble( &(pd->price),        pd->times,pd->paths,pd->underliers);
}


void pathParamsInit(pathData *pp){
  /* Uses pp->underliers, pp->covariance[i][j], pp->growthRate[i], and
   * pp-phrase to initialize the private data and the RANLIB library.
   */

  int i,j,k;

  /* Copy the 2d covariance matrix into the 1d scratch matrix */
  k = 0;
  for(i=0;i<pp->underliers;i++){
    for(j=0;j<pp->underliers;j++){
      pp->cov1[k] = pp->covariance[i][j];
      k++;
    }
  }

  /* Copy double growthRate array into float fgrowth array */
  /* Subtractthe dividend rate from the growth rate */
  for(i=0;i<pp->underliers;i++){
    pp->fgrowth[i] = pp->growthRate[i] - pp->dividendRate[i];
  }

  /* Give the data to RANLIB and let it do what it wants in parm */
  setgmn(pp->fgrowth,pp->cov1,pp->underliers,pp->parm);

  /* Hash the phrase into RNG seeds */
  phrtsd(pp->phrase,&(pp->seed1),&(pp->seed2));

  /* Initialize the RNG with the seeds */
  setall(pp->seed1,pp->seed2);

  /* Now whenever we want numbers from the distribution described in pp
   * we call genmn(pp->parm,vector,pp->cov1) where vector is a float array
   * of dimension [0:pp->underliers-1].  The random numbers go in vector
   */
}


void pathDataFill(pathData *pd){
  /* Uses pd->params, pd->paths, and pd->times to generate paths which are
   * placed in pd->price[p][t][u].  Make sure all these are initialized and
   * allocated first.  Each path starts with the value in pd->price[0][p][u]
   * so set these also.
   */

  int time, path, underlier,basePath;
  float *randomNum, std;

  randomNum = (float*)calloc(pd->underliers,sizeof(float));
  std = sqrt(pd->clusterVariance);

  /* pd->params describes the random numbers we get by calling
   * genmn(pd->params.parm, randomNum, pd->params.cov1).  To get the value in the next
   * time slice, we grow the values in the previous time slice by the random
   * numbers.  Each underlier gets its own growth rate.
   */

  for(time=1; time<pd->times; time++){
    for(path=0;path<pd->meshPaths;path++){
      if(path % pd->clusterSize == 0){
	genmn(pd->parm, randomNum, pd->cov1);
	for(underlier = 0; underlier < pd->underliers; underlier++){
	  pd->price[time][path][underlier] = (1+randomNum[underlier])*pd->price[time-1][path][underlier];
	}
	basePath = path;
      }else{
	for(underlier = 0; underlier < pd->underliers; underlier++){
	  pd->price[time][path][underlier] = pd->price[time][basePath][underlier] + gennor(0,std);
	}
      }
/*        printf("[%i][%i][%i] = ",time,path,underlier);  */
/*        for(underlier=0;underlier<pd->underliers;underlier++) printf("%lf, ",pd->price[time][path][underlier]);  */
/*        printf("\n");  */
    }
  }

  for(time=1; time<pd->times; time++){
    for(path=pd->meshPaths; path<pd->paths; path++){
      genmn(pd->parm, randomNum, pd->cov1);
      for(underlier = 0; underlier < pd->underliers; underlier++){
	pd->price[time][path][underlier] = (1+randomNum[underlier])*pd->price[time-1][path][underlier];
      }
    }
  }

  free(randomNum);
}


double pathTransitionProb(pathData *pp, int time, int initialPath, int finalPath){
  /* Calculates the probability of a transition from initial underlier proces
   * initialPrice[] to final underlier prices finalPrice[] over one timeslice.
   * Uses the statistics contained in pp.
   *
   * This is written for the case of independently distributed
   * underlier growth rates.  It should be expanded later.
   */

  int i;
  double exponential;
  static double *growth;
  static double constant = -1;

  /* Just calculate the constant the first time */
  if(constant == -1){
    /* Calculate the constant in front of the equation of the multivariate normal 
     * distribution (without corelations)
     */

    constant = pow(2*M_PI,pp->underliers);
    for(i=0; i<pp->underliers; i++){
      constant *= pp->covariance[i][i];
    }
    constant = 1.0/sqrt(constant);

    /* Allocate growth, same dimension as prices */
    allocate1dDouble(&growth,pp->underliers);
  }

  /* Find the growth rates corresponding to the price changes */
  /* Its the growth rates which are normally distributed */

  for(i=0; i<pp->underliers; i++){
    growth[i] = (pp->price[time+1][finalPath][i] - pp->price[time][initialPath][i])/
		 pp->price[time][initialPath][i];
  }

  /* Calculate the exponential term */
  exponential = 0;
  for(i=0; i<pp->underliers; i++){
    exponential += (growth[i] - pp->growthRate[i])*(growth[i] - pp->growthRate[i])/
      pp->covariance[i][i];
  }
  exponential = exp(-0.5*exponential);

  /* Put them together and we have our result */
  return(constant*exponential);
}
